#define _GNU_SOURCE
#include "config.h"

#include "include/ds485-socket-server.h"

#include <assert.h>
#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <poll.h>
#include <pthread.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <syslog.h>

#include <utlist.h>

#include "include/ds485-socket-interface.h"
#include "ds485-socket-server-impl.h"

#include "base/log.h"
#include "base/socket_server.h"
#include "base/util.h"

#ifndef POLLRDHUP
#define POLLRDHUP 0
#endif

#define SVCBUFLEN 8500

/**
 * How long server sleep until checking
 * for client connections again
 * will delay shutdown by this many seconds
 */
#define LISTEN_THREAD_SLEEP        10 /* sec */

static void ds485_server_process_client_data(dS485dClient *cl);
static void ds485_server_free_client(dS485dClient *cl);

static int update_pretty_name(dS485dClient *cl)
{
    struct sockaddr_in fsin;
    int rval;

    rval = get_peername(cl->fd, &fsin);
    if (rval < 0) {
        // already logged
        return rval;
    }

    /* increase peer_name size if doing changes */
    rval = snprintf(cl->peer_name, sizeof(cl->peer_name), "%.15s:%05d.%02d",
                    inet_ntoa(fsin.sin_addr),
                    ntohs(fsin.sin_port),
                    cl->transactionId);

    if (rval >= (int)sizeof(cl->peer_name)) {
        report(LOG_NOTICE, cl->peer_name, "snprintf truncated\n");
        /* minor issue, do not refuse client */
        return 0;
    }

    return 0;
}

/* CAUTION: requires cl->lock to be held */
void _ds485n_client_detach(dS485dClient *cl)
{
    if (-1 == cl->fd) {
        return;
    }

    report2(LOG_DEBUG, cl->peer_name, "detaching\n");
    close(cl->fd);
    cl->fd = -1;
    cl->ts_detach = time(NULL);
}

static void clDetach(dS485dClient* cl)
{
    pthread_mutex_lock(&cl->lock);   //xx
    _ds485n_client_detach(cl);
    pthread_mutex_unlock(&cl->lock);
}

/* CAUTION: clientListRWLock must be held */
static int _ds485_server_clients_count(struct dS485_server *server)
{
    dS485dClient* ign;
    int count = 0;

    LL_FOREACH(server->clSessions, ign) {
        count++;
    }
    return count;
}

static int ds485_server_append_client(struct dS485_server *server,
                                      dS485dClient* cl)
{
    pthread_rwlock_wrlock(&server->clientListRWLock);

    if (server->client_count >= server->client_limit) {
        report(LOG_INFO, "server: maximum number of clients (%d) "
               "reached\n", server->client_count);
        return -EUSERS;
    }

    if (server->client_count >= MAX_CLIENTS) {
        pthread_rwlock_unlock(&server->clientListRWLock);
        report2(LOG_INFO, cl->peer_name,
                "transaction id pool depleted, rejecting client\n");
        return -EBUSY;
    }

    if (_ds485_server_clients_count(server) != server->client_count) {
        pthread_rwlock_unlock(&server->clientListRWLock);
        report2(LOG_ERR, cl->peer_name,
                "append: client count mismatch.\n");
        return -ERANGE;
    }

    cl->server = server;
    cl->transactionId = server->transid_pool[server->client_count++];
    LL_PREPEND(server->clSessions, cl);
    pthread_rwlock_unlock(&server->clientListRWLock);

    (void)update_pretty_name(cl);
    return 0;
}

static int ds485_server_remove_client(struct dS485_server *server,
                                      dS485dClient* cl)
{
    pthread_rwlock_wrlock(&server->clientListRWLock);

    if (_ds485_server_clients_count(server) != server->client_count) {
        pthread_rwlock_unlock(&server->clientListRWLock);
        report(LOG_ERR, "remove: client count mismatch\n");
        return -ERANGE;
    }

    if (server->client_count <= 0) {
        pthread_rwlock_unlock(&server->clientListRWLock);
        report(LOG_ERR, "remove: client not registered\n");
        return -EINVAL;
    }

    /*
     * TODO: remove client is only called from the dispatch thread.
     * If we ever remove clients when message send fails, we need
     * a better check if the client has already been removed
     * from the list
     */

    server->transid_pool[--server->client_count] = cl->transactionId;
    LL_DELETE(server->clSessions, cl);
    pthread_rwlock_unlock(&server->clientListRWLock);

    return 0;
}

static void cleanup_helper(void *arg)
{
    dS485dClient *cl = arg;
    ds485_server_remove_client(cl->server, cl);
    ds485_server_free_client(cl);
}

static void* sysServiceThread(void* arg)
{
    dS485dClient *cl = arg;
    struct dS485_server *server = cl->server;
    struct pollfd fds[1];
    int rval;

    pthread_cleanup_push(cleanup_helper, cl);

    report2(LOG_NOTICE, cl->peer_name, "new client connection %d\n",
            cl->transactionId);

    fds[0].fd = cl->fd;
    fds[0].events = POLLIN;

    /* handle client connection */
    while (server->active && cl->fd > 0) {

        pthread_mutex_lock(&cl->lock);
        cl->ts_activity = time(NULL);
        pthread_mutex_unlock(&cl->lock);

        cl->m_state = CL_POLL;
        rval = poll(fds, 1, MAX_THREAD_SLEEP * 1000);
        if (rval < 0) {
            if (errno == EINTR) {
                continue;
            }
            report2(LOG_WARNING, cl->peer_name, "poll: %d [%d]\n", rval);
            break;
        } else if (rval == 0) {
            /* timeout */
            continue;
        }

        if (!server->active) {
            report2(LOG_INFO, cl->peer_name,
                    "poll: shutdown detected, disconnect client\n");
            break;
        }

        if (fds[0].revents & (POLLRDHUP | POLLERR | POLLHUP | POLLNVAL)) {
            report2(LOG_WARNING, cl->peer_name, "poll flags: 0x%04x\n", fds[0].revents);
            break;
        }

        assert(fds[0].revents == POLLIN);
        cl->m_state = CL_RECV;
        rval = recv(cl->fd, cl->rxBuffer + cl->rxLen, SVCBUFLEN - cl->rxLen, 0);
        if (rval < 0) {
            report2(LOG_WARNING, cl->peer_name, "recv: %s [errno %d]\n",
                    strerror(errno), errno);
            if (errno == ECONNRESET) {
                break;
            }
            continue;
        } else if (rval == 0) {
            report2(LOG_DEBUG, cl->peer_name, "peer closed connection\n");
            break;
        }

        cl->rxLen += rval;
        cl->m_state = CL_PROCESS;
        ds485_server_process_client_data(cl);
    }
    cl->m_state = CL_SHUTDOWN;

    /* pthread_cleanup_pop(true) is buggy on some platforms */
    pthread_cleanup_pop(0);

    cl->m_state = CL_CLOSED;
    report2(LOG_NOTICE, cl->peer_name, "connection terminated\n");
    cleanup_helper(cl); /* cl is removed */

    return NULL;
}

/**
 * ds485_server_init_client
 * @sock -- client socket
 */
static dS485dClient* ds485_server_create_client(int cfd)
{
    dS485dClient* cls;
    int opts, rval;

    opts = fcntl(cfd, F_GETFL);
    if (opts < 0) {
        report(LOG_WARNING, "new client: F_GETFL failed %s", strerror(errno));
        close(cfd);
        return NULL;
    }

    if (fcntl(cfd, F_SETFL, opts | O_NONBLOCK) < 0) {
        report(LOG_WARNING, "new client: setting O_NONBLOCK failed %s",
               strerror(errno));
        close(cfd);
        return NULL;
    }

    cls = calloc(sizeof(dS485dClient), 1);
    if (cls == NULL) {
        report(LOG_WARNING, "new client: %s", strerror(ENOMEM));
        close(cfd);
        return NULL;
    }

    cls->fd = cfd;
    cls->transactionId = -1; /* 0 is valid id */
    pthread_mutex_init(&cls->lock, NULL);

    rval = update_pretty_name(cls);
    if (rval) {
        report(LOG_ERR, "new client: failed to get peername %s(%d)\n",
               strerror(-rval), -rval);
        close(cfd);
        pthread_mutex_destroy(&cls->lock);
        free(cls);
        return NULL;
    }

    cls->rxBuffer = malloc(SVCBUFLEN);
    if (cls->rxBuffer == NULL) {
        report(LOG_ERR, "new client: failed to allocate recevice buffer\n");
        close(cfd);
        pthread_mutex_destroy(&cls->lock);
        free(cls);
        return NULL;
    }

    return cls;
}

static void ds485_server_free_client(dS485dClient *cl)
{
    clDetach(cl);
    pthread_mutex_destroy(&cl->lock);
    free(cl->rxBuffer);
    free(cl);
}

static void ds485_server_process_client_data(dS485dClient *cl)
{
    uint8_t  *p;

    p = cl->rxBuffer;
    while (cl->rxLen >= 3) {
        int pkt_len = p[2] << 8 | p[1];

        /* not enough data */
        if (pkt_len > cl->rxLen) {
            report2(LOG_DEBUG, cl->peer_name, "not enough data, retry (%d > %d)\n",
                    pkt_len, cl->rxLen);
            break;
        }

        /* Process packet data ... */
        ds485_netlib_packet_parser(cl, p);
        /* if (rval) TODO */

        p += pkt_len;
        cl->rxLen -= pkt_len;
    }

    if (p != cl->rxBuffer && cl->rxLen > 0) {
        memmove(cl->rxBuffer, p, cl->rxLen);
    }
}

static void *watchdog_thread(void *arg)
{
    struct dS485_server *server = arg;
    struct pollfd fds[1];
    time_t last = 0;
    int nfds, rval;

    fds[0].fd = server->wdctrl_pipe[0];
    fds[0].events = POLLIN;
    nfds = 1;

    while (server->active) {
        dS485dClient *cl, *blocked_thread = NULL;
        time_t now;

        rval = poll(fds, nfds, CLIENT_INACTIVE_THRESHOLD * 1000);
        if (rval < 0) {
            rval = errno;
            if (errno == EINTR) {
                continue;
            }
            /*
             * all other errors say; we are too
             * stupid to setup the pollfd array
             * TODO: panic, with strerror
             */
            report(LOG_WARNING, "poll thread: poll %s(%d)\n",
                   strerror(rval), rval);
            nfds = 0; /* fallback to polling */
            continue;
        }

        if (rval != 0) {
            /* Not a timeout, we woke up from the self-pipe */
            if (fds[0].revents & POLLERR) {
                report(LOG_WARNING, "watchdog control socket gone: poll 0x%04x\n",
                       fds[0].revents);
                break; /* give up */
            }

            if (!server->active) {
                /* this is the actual message, wake-up is only heads-up */
                break;
            }

            report(LOG_WARNING, "watchdog: unexpected wakeup from control socket\n");

            /*
             * This seems like an accidental write to our socket. Consume
             * the data and discard otherwise poll will never sleep again
             * because there is data to be read #15004
             */
            do {
                char buf[10];
                rval = read(fds[0].fd, buf, 10);
            } while (rval > 0);
        }

        time(&now);
        if (now - last < 10 * 60) { // 10 min
            continue;
        }
        last = now;

        pthread_rwlock_rdlock(&server->clientListRWLock);
        LL_FOREACH(server->clSessions, cl) {
            pthread_mutex_lock(&cl->lock);
            int is_blocked = now - cl->ts_activity > CLIENT_INACTIVE_THRESHOLD;
            pthread_mutex_unlock(&cl->lock);
            if (is_blocked) {
                blocked_thread = cl;
                break;
            }
        }

        if (!blocked_thread) {
            pthread_rwlock_unlock(&server->clientListRWLock);
            continue;
        }

        LL_FOREACH(server->clSessions, cl) {
            report2(LOG_WARNING, cl->peer_name,
                    "managed thread blocked:%d inactive:%4ds cmd:0x%02x fd:%d state:%d\n",
                    (now - cl->ts_activity > CLIENT_INACTIVE_THRESHOLD),
                    (now - cl->ts_activity), cl->last_command, cl->fd,
                    cl->m_state);
        }
        pthread_rwlock_unlock(&server->clientListRWLock);
    }

    report(LOG_NOTICE, "watchdog: terminating\n");
    return NULL;
}

static int start_watchdog(struct dS485_server* server)
{
    int rval = -1;

    rval = pipe2_wrapper(server->wdctrl_pipe);
    if (rval < 0) {
        report(LOG_WARNING, "failed to create watchdog control pipe %s(%d)\n",
               strerror(-rval), -rval);
        return rval;
    }

    rval = pthread_create(&server->watchdog_tid, NULL, watchdog_thread, server);
    if (rval) {
        report(LOG_ERR, "watchdog_thread start: %s\n", strerror(rval));
        close(server->wdctrl_pipe[0]);
        close(server->wdctrl_pipe[1]);
        return -rval;
    }

    return 0;
}

static int stop_watchdog(struct dS485_server* server)
{
    int rval, msg = 42; /* bogus value */

    rval = write(server->wdctrl_pipe[1], &msg, sizeof(msg));
    if (rval == -1) {
        rval = errno;
        report(LOG_WARNING, "failed stop watchdog thread: %s(%d)\n", strerror(rval),
               rval);
    }

    rval = pthread_join(server->watchdog_tid, NULL);
    if (rval) {
        report(LOG_WARNING, "failed to join watchdog thread: %s(%d)\n",
               strerror(rval), rval);

        rval = 0; /* ingore error */
    }

    /* pipes are valid if start_watchdog succeeds */
    close(server->wdctrl_pipe[0]);
    close(server->wdctrl_pipe[1]);
    return 0;
}

static int ds485_main_loop(struct dS485_server *server)
{
    struct pollfd fds[2];
    int rval, nfds;

    memset(&fds, 0, sizeof(fds));
    fds[0].fd = server->ctrl_pipe[0];
    fds[0].events = POLLIN;
    fds[1].fd = server->sfd;
    fds[1].events = POLLIN;
    nfds = 2;

    if (server->cb.status_change) {
        server->cb.status_change(server, -1);
    }

    while (true) {
        struct sockaddr fsin;
        socklen_t cnamelen;
        int cfd;

        /* socket is non-blocking, wait for client */
        rval = poll(fds, nfds, -1);
        if (rval < 0) {
            rval = errno;
            if (errno == EINTR) {
                continue;
            }
            report(LOG_ERR, "poll: %s\n", strerror(rval));
            return -rval;
        }

        if (fds[0].revents != 0) {
            assert(!server->active);
            break;
        }

        if (fds[1].revents & (POLLRDHUP | POLLERR | POLLHUP | POLLNVAL)) {
            /* TODO remove file descriptor, trigger action, continue */
            report(LOG_WARNING, "poll: file descriptor failed 0x%04x\n",
                   fds[1].revents);
            rval = -EBADF;
            break;
        }

        assert(fds[1].revents == POLLIN);
        cnamelen = sizeof(fsin);
        cfd = accept(server->sfd, (struct sockaddr *)&fsin, &cnamelen);
        if (cfd < 0) {
            report(LOG_WARNING, "new client: accept failed %s", strerror(errno));
            continue;
        }

        dS485dClient* cls = ds485_server_create_client(cfd);
        if (!cls) {
            continue;
        }

        rval = ds485_server_append_client(server, cls);
        if (rval < 0) {
            ds485_server_free_client(cls);
            if (rval == -EUSERS) {
                /* client limit reached */
                report(LOG_INFO, "server: sleep for %d secs\n", LISTEN_THREAD_SLEEP);
                sleep(LISTEN_THREAD_SLEEP);
            }
            continue;
        }

        /* start service thread */
        rval = pthread_create(&cls->tid, NULL, sysServiceThread, cls);
        if (rval) {
            report(LOG_ERR, "create client thread err:%s\n",
                   strerror(rval));
            ds485_server_remove_client(server, cls);
            ds485_server_free_client(cls);
            if (rval == EAGAIN) {
                rval = 0;
                continue; /* client will retry */
            }
            server->active = false;
            rval = -rval;
            break;
        }
        pthread_detach(cls->tid);
    }

    return 0;
}


NETLIB_EXPORTED int ds485_server_listen(struct dS485_server* server, int port, bool listenGlobal)
{
    return ds485_server_listen2(server, port, listenGlobal, (MAX_CLIENTS - 1));
}

NETLIB_EXPORTED int ds485_server_listen2(struct dS485_server* server, int port,
                                         bool listenGlobal, int maxClients)
{
    int rval = -1;

    if (!server->active) {
      /*
       * starting an ds485-server multiple times requires
       * locking the active/shutdown_complete flags
       * since !active could mean: shutting down, shutdown complete
       * or starting up but active not yet set to true
       * if you want to change this, pls make sure catch all possible
       * abort/listen2 combinations
       */
      report(LOG_ERR, "ERROR ds485_server can be started once only\n");
      return -EINVAL;
    }

    if (maxClients >= MAX_CLIENTS) {
        report(LOG_ERR, "ERROR maximum of clients exceeded. %d >= %d\n",
               maxClients, MAX_CLIENTS);
        return -EINVAL;
    } else if (maxClients <= 0) {
        server->client_limit = MAX_CLIENTS - 1;
    } else {
        server->client_limit = maxClients;
    }

    rval = passivesock(port, "tcp", 0, listenGlobal);
    if (rval < 0) {
        report(LOG_ERR, "ERROR opening network socket\n");
        return rval;
    }
    server->sfd = rval;

    rval = pipe2_wrapper(server->ctrl_pipe);
    if (rval < 0) {
        report(LOG_WARNING, "failed to create server control pipe %s(%d)\n",
               strerror(-rval), -rval);
        return rval;
    }

    rval = start_watchdog(server);
    if (rval) {
        report(LOG_ERR, "start watchdog failed\n");
        close(server->sfd);
        close(server->ctrl_pipe[0]);
        close(server->ctrl_pipe[1]);
        return rval;
    }

    rval = ds485_main_loop(server);

    report(LOG_INFO, "server: shutdown\n");
    stop_watchdog(server);
    close(server->sfd);
    close(server->ctrl_pipe[0]);
    close(server->ctrl_pipe[1]);
    server->shutdown_complete = true;
    return rval;
}

NETLIB_EXPORTED void ds485_server_abort(struct dS485_server *server)
{
    int ret, msg;
    report(LOG_NOTICE, "server: request shutdown\n");

    /* all threads periodically poll this variable */
    server->active = false;

    if (server->shutdown_complete) {
        /* ctrl pipe already closed */
        report(LOG_WARNING, "server: already stopped\n");
        return;
    }

    msg = 42; /* bogus value */
    ret = write(server->ctrl_pipe[1], &msg, sizeof(msg));
    if (ret == -1) {
        ret = errno;
        report(LOG_WARNING, "failed stop server: %s(%d)\n", strerror(ret), ret);
    }

    /* TODO pthread_join clients */
}
