#define _GNU_SOURCE
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <netinet/tcp.h>
#include <netinet/ip.h>
#include <poll.h>
#include <pthread.h>
#include <syslog.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <unistd.h>

#include <check.h>

#include "base/socket_send.h"
#include "base/socket_server.h"
#include "tests/lib_simple_client.h"

/**
 * export CFLAGS=--coverage
 * gcov -d base/.libs/libbase_la-socket_send.gcno
 * inspect socket_send.c.gcov to see if all branches taken
 */

pthread_t tid;
int controlfd;
int port;

typedef enum {
    c_setup,
    c_stop,
} cmd_t;

typedef int (*client_func_t)(int, void *);
struct control_msg {
    cmd_t cmd;
    int (*client_func)(int fd, void *);
    int client_events;
    void *client_arg;
};

typedef enum {
    e_none,
    e_ready,
    e_terminate,
} state_t;

struct server {
    struct pollfd *fds;
    int nfds;
    int (*client_func)(int fd, void *);
    short client_events;
    void *client_arg;

    state_t state;
    int listenfd;
    int controlfd;
};

void append_client(struct server *server, int sockfd, short events)
{
    server->fds = realloc(server->fds, (server->nfds + 1) *
                          sizeof(*server->fds));
    ck_assert_ptr_ne(server->fds, NULL);

    server->fds[server->nfds].fd = sockfd;
    server->fds[server->nfds].events = events;
    server->fds[server->nfds].revents = 0; /* valgrind */
    server->nfds++;
}

int remove_client(struct server *server, int sockfd)
{
    struct pollfd *t = NULL;
    int i;

    for (i = 0; i < server->nfds; i++) {
        if (server->fds[i].fd == sockfd) {
            t = &server->fds[i];
            break;
        }
    }

    if (!t) {
        return 0;
    }

    server->nfds--;
    for (; i < server->nfds; i++) {
        server->fds[i] = server->fds[i + 1];
    }

    server->fds = realloc(server->fds, server->nfds * sizeof(*server->fds));
    ck_assert_ptr_ne(server->fds, NULL);

    /* not implemented */
    return -EINVAL;
}

/*
 * TODO this should replace ds485_server_listen
 */
void *server_thread(void *arg)
{
    struct server *server = arg;
    int ret, i;

    printf("server_thread start\n");

    append_client(server, server->controlfd, POLLIN);
    append_client(server, server->listenfd, POLLIN);

    while (server->state != e_terminate) {

        ret = poll(server->fds, server->nfds, 0);
        if (ret == -1) {
            ret = errno;
            if (ret == EINTR) {
                continue;
            }
            perror("poll");
        }

        /* ret is number of active fd */
        for (i = 0; i < server->nfds; i++) {
            int fd = server->fds[i].fd;

            if (server->fds[i].revents == 0) {
                continue;
            }

            if (fd == server->controlfd) {
                struct control_msg msg;
                ret = read(fd, &msg, sizeof(msg));
                ck_assert_int_eq(ret, sizeof(msg));

                switch(msg.cmd) {
                case c_setup:
                    printf("server cmd: setup\n");
                    server->client_func = msg.client_func;
                    server->client_events = msg.client_events;
                    server->client_arg = msg.client_arg;
                    server->state = e_ready;
                    break;
                case c_stop:
                    printf("server cmd: stop\n");
                    server->state = e_terminate;
                    break;
                }

            } else if (fd == server->listenfd) {
                /* new client */
                ck_assert_int_eq(server->state, e_ready);
                printf("server new client\n");
                int cl = accept(server->listenfd, NULL, NULL);
                if (cl < 0) {
                    continue;
                }

                /* ensure tcp_nodelay is inherited from listen socket */
                socklen_t tmp, len = sizeof(tmp);
                ret = getsockopt(cl, IPPROTO_TCP, TCP_NODELAY, &tmp, &len);
                ck_assert_int_eq(ret, 0);
                ck_assert_int_eq(tmp, 1);

                append_client(server, cl, server->client_events);

            } else {
                /* must be client */
                ck_assert_int_ne(server->state, e_none);
                ck_assert_ptr_ne(server->client_func, NULL);
                ret = server->client_func(fd, server->client_arg);
                if (!ret) {
                    /* disconnect client */
                    close(fd);
                    remove_client(server, fd);
                }
            }
        }
    }

    for (i = 0; i < server->nfds; i++) {
        close(server->fds[i].fd);
    }
    free(server->fds);
    free(server);

    printf("server terminates\n");
    return NULL;
}

void start_server(void)
{
    int ret;
    struct server *server;
    int _controlfd[2];

    server = calloc(sizeof(struct server), 1);
    ck_assert_ptr_ne(server, 0);

    ret = pipe2(_controlfd, O_DIRECT);
    if (ret == -1) {
        /* O_DIRECT since kernel 3.4 */
        ret = pipe(_controlfd);
    }
    ck_assert_int_eq(ret, 0);
    server->controlfd = _controlfd[0];
    controlfd = _controlfd[1];

    port = 5572;
    ret = passivesock(port, "tcp", 2, false);
    if (ret < 0) {
        ck_abort_msg("failed to create server port");
    }
    server->listenfd = ret;

    ret = pthread_create(&tid, NULL, server_thread, server);
    if (ret < 0) {
        ck_abort_msg("failed to create server");
    }
}

void teardown_server(void)
{
    struct control_msg msg;
    int ret;

    memset(&msg, 0, sizeof(msg));
    msg.cmd = c_stop;
    ret = write(controlfd, &msg, sizeof(msg));
    ck_assert_int_eq(ret, sizeof(msg));
    pthread_join(tid, NULL);
}

/*
 * test cases
 */

void prepare_server(client_func_t client_func, short client_events,
                    void *data)
{
    struct control_msg msg;
    int ret;

    memset(&msg, 0, sizeof(msg));
    msg.cmd = c_setup;
    msg.client_func = client_func;
    msg.client_events = client_events;
    msg.client_arg = data;
    ret = write(controlfd, &msg, sizeof(msg));
    ck_assert_int_eq(ret, sizeof(msg));
}

/* buffer size will affect speed of unit-test */
char buf[512 * 1024];
int discard(int fd, void __attribute((unused)) *data)
{
    ssize_t ret;

    ret = read(fd, buf, sizeof(buf));
    ck_assert_int_ne(ret, -1);
    if (ret == 0) {
        printf("peer closed\n");
        return false;
    }
    return true;
}

START_TEST (test_socket_send)
{
#define SZ (1024 * 1024 * 10)
    unsigned char* buf;
    int i, s, ret;

    buf = malloc(SZ);
    ck_assert_ptr_ne(buf, NULL);
    memset(buf, 0x7f, SZ);

    prepare_server(discard, POLLIN, NULL);

    s = connect_server(port);
    ck_assert_int_ge(s, 0);

    for (i = 0; i < 10; i++) {
        ret = socket_send(s, buf, SZ);
        ck_assert_int_eq(ret, 0);
    }
#undef SZ
    free(buf);
    close(s);
}
END_TEST


struct sendv_data {
    struct iovec *iov;
    int iovct;
    int repetition;
};

int sendv_sender(int fd, void *arg)
{
    struct sendv_data *d = arg;
    int ret;

    ret = socket_sendv(fd, d->iov, d->iovct);
    ck_assert_int_eq(ret, 0);

    return (--d->repetition > 0);
};

START_TEST (test_socket_sendv_2big)
{
    unsigned char *snd_buf;
    struct iovec *iov;
    int ret, s, i;

#define SZ_IOV 1025
#define SZ_BUF 1024

    /*
     * change in kernel since 31a25fae85956
     * New behavior is that kernel will truncate transfers MAX_RW_COUNT
     * typically INT_MAX - 4095 max. Previously bigger transfer
     * produced an error.
     */
    /* 1024 iov max */
    iov = calloc(sizeof(*iov), SZ_IOV);
    ck_assert_ptr_ne(iov, NULL);

    snd_buf = malloc(SZ_BUF);
    ck_assert_ptr_ne(iov, NULL);

    for (i = 0; i < SZ_IOV; i++) {
        iov[i].iov_base = snd_buf;
        iov[i].iov_len = SZ_BUF;
    }

    prepare_server(discard, POLLOUT, NULL);

    s = connect_server(port);
    ck_assert_int_ge(s, 0);

    ret = socket_sendv(s, iov, SZ_IOV);
    ck_assert_int_eq(ret, -EMSGSIZE);

#undef SZ_BUF
#define SZ_BUF (1024 * 1024 * 2)

    /* 2024 MB(INT_MAX + 1) messages will truncate  */
    snd_buf = realloc(snd_buf, SZ_BUF);
    for (i = 0; i < 1024; i++) {
        iov[i].iov_base = snd_buf;
        iov[i].iov_len = SZ_BUF;
    }

    ret = socket_sendv(s, iov, 1024);
    ck_assert_int_eq(ret, -EMSGSIZE);

#undef SZ_IOV
#undef SZ_BUF

    free(snd_buf);
    free(iov);
    close(s);
}
END_TEST

START_TEST (test_socket_sendv)
{
    /*
     * ensure socket_sendv either refuses the messages
     * or processes it without truncation
     * INT_MAX is the limit which translates to 2 GB - 1B
     */
#define SZ (1024 * 1024 * 2)
#define IOVCNT 1024

    struct sendv_data data;
    unsigned char *snd_buf, *recv_buf;
    ssize_t ret, recv_bytes, total;
    int s, i;

    snd_buf = malloc(SZ);
    ck_assert_ptr_ne(snd_buf, NULL);
    memset(buf, 0x7f, 1024);

    data.iov = calloc(sizeof (struct iovec), IOVCNT);
    ck_assert_ptr_ne(data.iov, NULL);
    data.iovct = IOVCNT;

    for (i = 0; i < IOVCNT; i++) {
        data.iov[i].iov_base = snd_buf;
        data.iov[i].iov_len = SZ;
    }
    data.iov[IOVCNT - 1].iov_len -= 4096; /* minus 4 kByte */

    total = 0;
    for (i = 0; i < IOVCNT; i++) {
        total += data.iov[i].iov_len;
    }
    ck_assert_int_eq(total, INT_MAX - 4095);

    prepare_server(sendv_sender, POLLOUT, &data);

    recv_buf = malloc(SZ);
    ck_assert_ptr_ne(data.iov, NULL);

    s = connect_server(port);
    ck_assert_int_ge(s, 0);

    recv_bytes = 0;
    do {
        ret = recv(s, recv_buf, SZ, MSG_NOSIGNAL);
        ck_assert_int_ne(ret, -1);
        recv_bytes += ret;
    } while (ret != 0);

    ck_assert_int_eq(recv_bytes, total);

#undef SZ
#undef IOVCNT

    free(snd_buf);
    free(recv_buf);
    free(data.iov);
    close(s);
}
END_TEST

/**
 * send as couple of more typically sized messages
 */
START_TEST (test_low_latency)
{
    unsigned char *hdr, *body, *recv_buf;
    struct sendv_data d;
    struct iovec *iov;
    int s, ret;

    /*
     * no tcp_nodelay/tcpnodlay: 12.8ms/9.5ms         | 1094/475 pkts
     * only tcp_nodelay        : 19.4ms/19.3ms/12.4ms | 2054/2024/799 pkts
     * only tos_lowdelay       : 9.6ms/6.9ms/14.2ms   | 454/172/1402 pkts
     * tcp_nodelay/tos_lowdelay: 20ms/18ms/12.2ms     | 2054/2054/792 pkt
     *
     *  since nagle would delay consecutive message in the
     *  same direction if peer is not sending messages
     *  earlier
     * -> nagle algorithm probably not active on localhost
     */

    iov = calloc(sizeof(struct iovec), 2);
    ck_assert_ptr_ne(iov, NULL);

    hdr = malloc(512);
    ck_assert_ptr_ne(hdr, NULL);
    memset(hdr, 0xf7, 64);

    body = malloc(512);
    ck_assert_ptr_ne(body, NULL);
    memset(body, 0x7f, 512);

    iov[0].iov_base = hdr;
    iov[0].iov_len = 64;
    iov[1].iov_base = body;
    iov[1].iov_len = 512;

    d.iov = iov;
    d.iovct = 2;
    d.repetition = 1024;

    prepare_server(sendv_sender, POLLOUT, &d);

    recv_buf = malloc(1024 * 1024);
    ck_assert_ptr_ne(recv_buf, NULL);

    s = connect_server(port);
    ck_assert_int_gt(s, 0);

    do {
        ret = recv(s, recv_buf, 1024 * 1024, MSG_NOSIGNAL);
        ck_assert_int_ne(ret, -1);
    } while(ret != 0);

    free(d.iov);
    free(body);
    free(hdr);
    close(s);
}
END_TEST

Suite *f_suite()
{
    Suite *s;
    TCase *tc_socket_send;

    s = suite_create("ds485n");
    tc_socket_send = tcase_create("socket_send");

    tcase_add_checked_fixture(tc_socket_send, start_server, teardown_server);

    tcase_set_timeout(tc_socket_send, 120);
    tcase_add_test(tc_socket_send, test_socket_send);
    tcase_add_test(tc_socket_send, test_socket_sendv);
    tcase_add_test(tc_socket_send, test_socket_sendv_2big);
    tcase_add_test(tc_socket_send, test_low_latency);
    suite_add_tcase(s, tc_socket_send);

    return s;
}

int main()
{
    Suite *s;
    SRunner *sr;
    int failed;

    s = f_suite();
    sr = srunner_create(s);

    srunner_run_all(sr, CK_VERBOSE);
    failed = srunner_ntests_failed(sr);
    srunner_free(sr);

    return (failed == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
}
